/*
** $Id: modlua_lua.c,v 1.1.1.1 2003/06/30 19:18:25 michel Exp $
** mod_lua - Apache module for Lua language
** See copyright notice in distro's COPYRIGHT file
*/

#include <string.h>
#include "modlua_lua.h"
#include <http_log.h>
#include <lualib.h>
#include <lauxlib.h>

static lua_State *modlua_lua_L = NULL;

typedef struct {
    void *r;
} modlua_lua_request_t;

static void modlua_lua_cleanup_state(void *v)
{
    lua_close((lua_State *)v);
}

#define REQUEST_TYPE "8b2f4802-2b30-4e47-8dd8-d70e2cadf6d4"

static request_rec *checkrequest(lua_State *L)
{
    modlua_lua_request_t *ud = (modlua_lua_request_t *)luaL_checkudata(L, 1,
        REQUEST_TYPE);
    luaL_argcheck(L, ud != NULL, 1, "'request' expected");
    if (ud == NULL) luaL_error(L, "attempt to use a closed request");
    return ud->r;
}

static int request_add_cgi_vars(lua_State *L)
{
    request_rec *r = checkrequest(L);
    ap_add_cgi_vars(r);
    return 0;
}

static int request_add_common_vars(lua_State *L)
{
    request_rec *r = checkrequest(L);
    ap_add_common_vars(r);
    return 0;
}

static int request_block_alarms(lua_State *L)
{
    ap_block_alarms();
}

static int request_content_type(lua_State *L)
{
    request_rec *r = checkrequest(L);
    if (lua_gettop(L) > 1) {
        char *s = ap_pstrdup(r->pool, luaL_checkstring(L, 2));
        ap_content_type_tolower(s);
        r->content_type = s;
    }
    
    lua_pushstring(L, r->content_type);
    return 1;
}

static int request_fileexists(lua_State *L)
{
    request_rec *r = checkrequest(L);
    lua_pushboolean(L, r->finfo.st_mode != 0);
    return 1;
}

static int request_filename(lua_State *L)
{
    request_rec *r = checkrequest(L);
    lua_pushstring(L, r->filename);
    return 1;
}

static int request_getenv(lua_State *L)
{
    request_rec *r = checkrequest(L);
    const char *s = luaL_checkstring(L, 2);
    lua_pushstring(L, ap_table_get(r->subprocess_env, s));
    return 1;
}

static int request_get_client_block(lua_State *L)
{
    request_rec *r = checkrequest(L);
    size_t rlen;  /* how much to read */
    size_t nr;  /* number of chars actually read */
    size_t n = (size_t)luaL_checknumber(L, 2);
    luaL_Buffer b;

    if (!n) {
        lua_pushliteral(L, "");
        return 1;
    }

    luaL_buffinit(L, &b);
    rlen = LUAL_BUFFERSIZE;  /* try to read that much each time */
    do {
        char *p = luaL_prepbuffer(&b);
        if (rlen > n) rlen = n;  /* cannot read more than asked */
        nr = ap_get_client_block(r, p, rlen);
        luaL_addsize(&b, nr);
        n -= nr;  /* still have to read `n' chars */
    } while (n > 0 && nr == rlen);  /* until end of count or eof */
    luaL_pushresult(&b);  /* close buffer */
    return 1;
}

static int request_hard_timeout(lua_State *L)
{
    request_rec *r = checkrequest(L);
    ap_hard_timeout(luaL_checkstring(L, 2), r);
}

static int request_header_only(lua_State *L)
{
    request_rec *r = checkrequest(L);
    lua_pushboolean(L, r->header_only);
    return 1;
}

static int request_headers_out(lua_State *L)
{
    request_rec *r = checkrequest(L);
    const char *key = luaL_checkstring(L, 2);
    if (lua_gettop(L) > 2) {
        const char *value = luaL_checkstring(L, 3);
        ap_table_set(r->headers_out, key, value);
        lua_settop(L, 3);
        return 1;
    }
    
    lua_pushstring(L, ap_table_get(r->headers_out, key));
    return 1;
}

static int request_kill_timeout(lua_State *L)
{
    request_rec *r = checkrequest(L);
    ap_kill_timeout(r);
}

static int request_log(lua_State *L)
{
    request_rec *r = checkrequest(L);
    const char *s = luaL_checkstring(L, 2);
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO | APLOG_WARNING, r, "%s", s);
    return 0;
}

static int request_remaining(lua_State *L)
{
    request_rec *r = checkrequest(L);
    lua_pushnumber(L, r->remaining);
    return 1;
}

static int request_reset_timeout(lua_State *L)
{
    request_rec *r = checkrequest(L);
    ap_reset_timeout(r);
}

static int request_rwrite(lua_State *L)
{
    request_rec *r = checkrequest(L);
    int nargs = lua_gettop(L);
    int count = 0;
    int i;

    for (i = 2; i <= nargs; i++) {
        int c = ap_rwrite(luaL_checkstring(L, i), lua_strlen(L, i), r);
        if (c == -1) {
            count = -1;
            break;
        }
        count += c;
    }
   
    lua_pushnumber(L, count);
    return 1;
}

static int request_send_http_header(lua_State *L)
{
    request_rec *r = checkrequest(L);
    ap_send_http_header(r);
    return 0;
}

static int request_setup_client_block(lua_State *L)
{
    request_rec *r = checkrequest(L);
    lua_pushnumber(L, ap_setup_client_block(r, luaL_checkint(L, 2)));
    return 1;
}

static int request_should_client_block(lua_State *L)
{
    request_rec *r = checkrequest(L);
    lua_pushboolean(L, ap_should_client_block(r));
    return 1;
}

static int request_soft_timeout(lua_State *L)
{
    request_rec *r = checkrequest(L);
    ap_soft_timeout(luaL_checkstring(L, 2), r);
}

static int request_status(lua_State *L)
{
    request_rec *r = checkrequest(L);
    if (lua_gettop(L) > 1) {
        const int status = luaL_checkint(L, 2);
        r->status = status;
        r->status_line = ap_psprintf(r->pool, "%i", status);
    }
    
    lua_pushnumber(L, r->status);
    return 1;
}

static int request_unblock_alarms(lua_State *L)
{
    ap_unblock_alarms();
}

static const luaL_reg request_m[] = {
    {"add_cgi_vars",            request_add_cgi_vars},
    {"add_common_vars",         request_add_common_vars},
    {"block_alarms",            request_block_alarms},
    {"content_type",            request_content_type},
    {"fileexists",              request_fileexists},
    {"filename",                request_filename},
    {"getenv",                  request_getenv},
    {"get_client_block",        request_get_client_block},
    {"hard_timeout",            request_hard_timeout},
    {"header_only",             request_header_only},
    {"headers_out",             request_headers_out},
    {"kill_timeout",            request_kill_timeout},
    {"log",                     request_log},
    {"remaining",               request_remaining},
    {"reset_timeout",           request_reset_timeout},
    {"rwrite",                  request_rwrite},
    {"send_http_header",        request_send_http_header},
    {"setup_client_block",      request_setup_client_block},
    {"should_client_block",     request_should_client_block},
    {"soft_timeout",            request_soft_timeout},
    {"status",                  request_status},
    {"unblock_alarms",          request_unblock_alarms},
    {NULL, NULL}
};

/* This function can to be called many times, but global lua_State just
 * will create one time. */
void modlua_lua_init(pool *p)
{
    lua_State *L;
            
    if (modlua_lua_L)
        return;

    L = lua_open();
    modlua_lua_L = L;
    ap_register_cleanup(p, L, modlua_lua_cleanup_state, ap_null_cleanup);

    luaopen_base(L);
    luaopen_table(L);
    luaopen_string(L);
    luaopen_math(L);
    luaopen_io(L);
    luaopen_loadlib(L);
              
    /* userdata REQUEST_TYPE */
    luaL_newmetatable(L, REQUEST_TYPE);
    lua_pushstring(L , "__index");
    lua_pushvalue(L, -2);
    lua_settable(L, -3);
    luaL_openlib(L, NULL, request_m, 0);
}

void modlua_lua_setluapath(const char *lua_path)
{
    lua_State *L = modlua_lua_L;
    lua_pushstring(L, lua_path);
    lua_setglobal(L, "LUA_PATH");
}

/* Call the Lua function require(filename) function.
 * Before call require, set LUA_PATH with modlua_lua_setluapath.
 * NOTE: the pool is used to temporary operations. */
const char *modlua_lua_require(pool *p, const char *filename)
{
    lua_State *L = modlua_lua_L;
    char *result = NULL;

    /* require(filename) */
    lua_getglobal(L, "require");
    if (!lua_isfunction(L, -1)) {
        lua_pop(L, 1);
        result = "The require global is not a function";
    }
    else {
        lua_pushstring(L, filename);
        if (lua_pcall(L, 1, 0, 0))
            result = ap_pstrcat(p, "The pcall to require(\"", filename,
                "\") fail: ", lua_tostring(L, -1), NULL);
    }

    return result;        
}

static void modlua_lua_cleanup_request(void *v)
{
    lua_State *L = modlua_lua_L;
    int i = (int)v;
    modlua_lua_request_t *lua_r;
    
    /* Set NULL in userdatum and remove reference from registry. */
    lua_rawgeti(L, LUA_REGISTRYINDEX, i);
    lua_r = (modlua_lua_request_t *)lua_touserdata(L, -1);
    lua_r->r = NULL;
    lua_pop(L, 1);
    luaL_unref(L, LUA_REGISTRYINDEX, i);
}

/* Push on Lua stack a userdatum with request_rec. */
static void pushrequest(lua_State *L, request_rec *r)
{
    modlua_lua_request_t *lua_r;
    int i;

    /* Get a userdatum for request_rec. */
    lua_r = (modlua_lua_request_t *)lua_newuserdata(L,
        sizeof(modlua_lua_request_t));
    luaL_getmetatable(L, REQUEST_TYPE);
    lua_setmetatable(L, -2); 

    lua_r->r = r;

    /* Store the userdatum in registry for use on cleanup. */
    lua_pushvalue(L, -1);
    i = luaL_ref(L, LUA_REGISTRYINDEX);
    ap_register_cleanup(r->pool, (void *)i, modlua_lua_cleanup_request,
        ap_null_cleanup);
}

/* Get function from funcname and push on Lua stack.
 * On sucess return true and push function on Lua stack.
 * On fail return false and push nothing. */
static int getfunc(lua_State *L, const char *funcname)
{
    const char *package;
    const char *p;
    int done = 0;
    
    if (!(package && *package))
        return done;

    package = funcname;
    lua_pushvalue(L, LUA_GLOBALSINDEX);
    do {
        if (!lua_istable(L, -1))
            break;

        p = package;
        while(*p && *p != '.') p++;
        lua_pushlstring(L, package, p - package);
        lua_gettable(L, -2);
        lua_remove(L, -2);

        package = *p ? p + 1 : p;
    } while (*package);

    if (lua_istable(L, -1)) {
        lua_pushstring(L, "handler");
        lua_gettable(L, -2);
    }                

    if (lua_isfunction(L, -1)) {
        done = 1;
    }
    else {
        lua_pop(L, 1);
    }

    return done;
}

/* return true if funcname is a function name valid. */
int modlua_lua_isfuncname(const char *funcname)
{
    const char *p;

    if (!funcname)
        return 0;

    /* funcname =~ /^[a-zA-Z_][0-9a-zA-Z_]*(\.[a-zA-Z_][0-9a-zA-Z_]*)*$/ */
    p = funcname;
    /* ^[a-zA-Z_] */
    if (*p >= 'a' && *p <= 'z' || *p >= 'A' && *p <= 'Z' || *p == '_')
        p++;
    else
        return 0;
    /* [0-9a-zA-Z_]* */
    while (*p >= '0' && *p <= '9' || *p >= 'a' && *p <= 'z' ||
        *p >= 'A' && *p <= 'Z' || *p == '_')
        p++;
    /* (\.[a-zA-Z_][0-9a-zA-Z_]*)* */
    if (*p == '.') {
        p++;
        
        /* [a-zA-Z_] */
        if (*p >= 'a' && *p <= 'z' || *p >= 'A' && *p <= 'Z' || *p == '_')
            p++;
        else
            return 0;

        /* [0-9a-zA-Z_]* */
        while (*p >= '0' && *p <= '9' || *p >= 'a' && *p <= 'z' ||
            *p >= 'A' && *p <= 'Z' || *p == '_')
            p++;
    }
    /* $ */
    if (*p) return 0;

    return 1;
}

/* Call the Lua function funcname(r)
 * The errmsg parameter always get value. When there is not message,
 * *errmsg = "". */
int modlua_lua_call_handler(request_rec *r, const char *funcname,
                            const char **errmsg)
{
    lua_State *L = modlua_lua_L;
    int status = HTTP_INTERNAL_SERVER_ERROR; 

    *errmsg = "";
    
    if (!getfunc(L, funcname)) {
        *errmsg = "Lua handler is unknown, likely a typo or "
            "Lua data corrupted or the package required didn't load";
        goto FINISH;
    }

    pushrequest(L, r);
    
    /* Call func(r) */
    if (lua_pcall(L, 1, 1, 0) != 0) {
        *errmsg = ap_pstrdup(r->pool, lua_tostring(L, -1));
        goto FINISH;
    }

    /* Get the return. */
    if (!lua_isnumber(L, -1)) {
        *errmsg = "The handler didn't return a number";
        goto FINISH;
    }
    status = lua_tonumber(L, -1);
    
FINISH:
    lua_settop(L, 0);
    return status;
}
